{-# LANGUAGE LambdaCase        #-}
{-# LANGUAGE OverloadedStrings #-}


module Datact
  ( datactMain
  ) where


import Control.Lens hiding (argument)
import Control.Exception.Lens
import Control.Monad.Reader

import qualified Data.Map.Strict as Map
import Data.Maybe
import qualified Data.Text.Lazy.IO as LIO
import qualified Data.Text.Lazy as Text

import Options.Applicative

import System.Exit
import System.IO
import System.IO.Error.Lens

import Text.PrettyPrint.Leijen.Text (pretty, renderPretty, displayT)
import Text.Read (readMaybe)


import Datact.Actions
import qualified Datact.Constants as Constants
import qualified Datact.SymbolTable as SymbolTable
import Datact.Parser
import Datact.Syntax


data Options = Options
  { sourceFile       :: FilePath
  , defaultRange     :: Int
  , rangeFile        :: Maybe FilePath
  , outputFile       :: Maybe FilePath
  , removeFromModule :: Maybe String
  }


type App = ReaderT Options IO


runApp :: App a -> Options -> IO a
runApp = runReaderT


datactMain :: IO ()
datactMain = handling _IOException ioeHandler $
    execParser options >>= runApp datact
  where
    options      = info (helper <*> datactOptions) mempty
    ioeHandler e = do
        let file = fromMaybe "<unknown source>" $ e^.fileName
        hPutStrLn stderr $ e^.description ++ ": " ++ file
        exitWith $ ExitFailure 2


datactOptions :: Parser Options
datactOptions = Options
    <$> strArgument (metavar "<model>")
    <*> argument auto (metavar "<default-range>")
    <*> optional (strArgument (metavar "<range-file>"))
    <*> optional (strOption (short 'o'))
    <*> optional (strOption (long "remove-in"))


datact :: App ()
datact = do
    modelPath <- asks sourceFile
    modelSrc  <- liftIO $ LIO.readFile modelPath

    case parseModel modelPath modelSrc of
        Left err -> liftIO $ print err
        Right model -> do
            upper      <- asks defaultRange
            removeFrom <- fmap Text.pack <$> asks removeFromModule

            actRanges <- readRangeFile

            let model' = expandModel (Range 1 upper) actRanges removeFrom model
                modelSrc' = displayT (renderPretty 1.0 300 (pretty model'))

            asks outputFile >>= liftIO . \case
                Just outputPath -> LIO.writeFile outputPath modelSrc'
                Nothing -> LIO.hPutStr stdout modelSrc'


readRangeFile :: App Actions
readRangeFile = asks rangeFile >>= \case
    Just path -> do
        rangeFileSrc <- liftIO $ LIO.readFile path
        return (toActionRanges rangeFileSrc)
    Nothing -> return Map.empty
  where
    toActionRanges =
        Map.fromList .
        mapMaybe (toActionRange . fmap Text.strip . Text.splitOn ",") .
        Text.lines
    toActionRange [act, r] = case readMaybe (Text.unpack r) of
        Just i  -> Just (act, Range 1 i)
        Nothing -> Nothing
    toActionRange _ = Nothing


expandModel :: Range -> Actions -> Maybe Name -> LModel -> Model ()
expandModel range actRanges removeFrom model =
    let Model defs = () <$ model
        consts     = Constants.fromModel model
        symTbl     = SymbolTable.fromModel consts model
        actRanges' = propagate actRanges model
        defs'      =
            over (traverse._ModuleDef)
            (expandModule' consts symTbl actRanges') defs
        defs'' = over (traverse._RenameDef) (expandRenaming actRanges') defs'
        defs''' =
            over (traverse._RewardsDef) (expandRewards range actRanges') defs''
    in Model defs'''
  where
    expandModule' consts symTbl actRanges' m =
        let opt = if Just (modName m) == removeFrom then Remove else Minimize
        in expandModule consts symTbl range actRanges' opt m
