{-# LANGUAGE OverloadedStrings #-}


module Datact.Actions
  ( Actions
  , Range(..)

  , Optimization(..)
  , propagate
  , expandModule
  , expandRenaming
  , expandRewards
  ) where


import Control.Arrow ((&&&))
import Control.Lens

import           Data.Map.Strict (Map)
import qualified Data.Map.Strict as Map
import           Data.Maybe
import           Data.Monoid

import qualified Data.Text.Lazy as Text


import Datact.Constants
import Datact.Guard
import Datact.SymbolTable
import Datact.Syntax


type Actions = Map Name Range


propagate :: Actions -> Model a -> Actions
propagate acts (Model defs) = foldr insert acts renamings
  where
    renamings = toListOf (traverse._RenameDef.to rnRenamings.traverse) defs
    insert (orig, new) acts' = case Map.lookup orig acts' of
        Just range -> Map.insert new range acts'
        Nothing    -> acts'


expandModule ::
       Constants
    -> SymbolTable
    -> Range
    -> Actions
    -> Optimization
    -> Module a
    -> Module ()
expandModule consts symTbl defaultRange actRanges opt (Module name isTemplate vars stmts _) =
    let vars' = fmap (() <$) vars
        stmts' = concatMap optimizedExpand stmts
    in Module name isTemplate vars' stmts' ()
  where
    optimizedExpand =
        mapMaybe (optimize consts symTbl opt name) .
        expandStmt consts defaultRange actRanges


expandStmt :: Constants -> Range -> Actions -> Stmt a -> [Stmt ()]
expandStmt consts defaultRange actRanges (Stmt (ActionSet acts ty) grd upds _) =
    go acts
  where
    go (act : acts') = do
        i <- lookupRange defaultRange actRanges act
        Stmt (ActionSet iActs _) iGrd iUpds _ <- go acts'
        let subst  = substitute act (intExpr (fromIntegral i))
            iGrd'  = removeTrueClauses consts (subst iGrd)
            iUpds' = fmap (over exprs subst) iUpds
            iActs' = indexedAct act i : iActs
        return (Stmt (ActionSet iActs' ty) iGrd' iUpds' ())
    go [] = [Stmt (ActionSet [] ty) (() <$ grd) (fmap (() <$) upds) ()]


expandRenaming :: Actions -> Renaming a -> Renaming a
expandRenaming actRanges rn =
    rn { rnRenamings = concatMap expandActionRenaming (rnRenamings rn) }
  where
    expandActionRenaming (orig, new) =
        case Map.lookup orig actRanges of
            Just (Range lower upper) ->
                fmap (indexedAct orig &&& indexedAct new) [lower .. upper]
            Nothing -> [(orig, new)]


expandRewards :: Range -> Actions -> Rewards a -> Rewards ()
expandRewards defaultRange actRanges (Rewards name rews _) =
    let rews' = concatMap (expandReward defaultRange actRanges) rews
    in Rewards name rews' ()


expandReward :: Range -> Actions -> Reward a -> [Reward ()]
expandReward defaultRange actRanges (Reward (Just (ActionSet acts ty)) grd rew _) =
    go acts
  where
    go (act : acts') = do
        i <- lookupRange defaultRange actRanges act
        Reward (Just (ActionSet iActs _)) iGrd iRew _ <- go acts'
        let subst  = substitute act (intExpr (fromIntegral i))
            iGrd'  = subst iGrd
            iRew'  = subst iRew
            iActs' = indexedAct act i : iActs
        return (Reward (Just (ActionSet iActs' ty)) iGrd' iRew' ())
    go [] = [Reward (Just (ActionSet [] ty)) (() <$ grd) (() <$ rew) ()]

expandReward _ _ r = [() <$ r]


lookupRange :: Range -> Actions -> Name -> [Int]
lookupRange defaultRange actRanges act = case Map.lookup act actRanges of
    Just range -> toList range
    Nothing    -> toList defaultRange


data Optimization
  = Remove
  | Minimize


optimize ::
       Constants
    -> SymbolTable
    -> Optimization
    -> Name
    -> Stmt ()
    -> Maybe (Stmt ())
optimize consts symTbl opt moduleName stmt
  | isFalse consts symTbl moduleName (stmtGuard stmt) = case opt of
        Remove   -> Nothing
        Minimize -> Just (Stmt (stmtAction stmt) (BoolExpr False ()) [] ())
  | otherwise = Just stmt


removeTrueClauses :: Constants -> Expr () -> Expr ()
removeTrueClauses consts =
    view conjunction . filter (not . isTrue) . view (from conjunction)
  where
    isTrue e = case analyzeClosedClause consts e of
        Just True -> True
        _         -> False


toList :: Range -> [Int]
toList (Range lower upper) = [lower .. upper]


indexedAct :: Name -> Int -> Name
indexedAct act i = act <> "_" <> Text.pack (show i)
