{-# LANGUAGE LambdaCase #-}


module Datact.Guard
  ( isFalse
  , analyzeClosedClause
  ) where


import Control.Lens

import           Data.Map.Strict (Map)
import qualified Data.Map.Strict as Map
import           Data.Maybe      (mapMaybe)
import           Data.Set        (Set)
import qualified Data.Set        as Set


import Datact.Constants
import Datact.Eval
import Datact.SymbolTable
import Datact.Syntax


isFalse :: Constants -> SymbolTable -> Name -> Expr a -> Bool
isFalse consts symTbl moduleName e = case result of
    ResultFalse -> True
    _           -> False
  where
    clauses = view (from conjunction) (() <$ e)
    result  = normalize $
        mconcat (fmap (analyzeClause consts symTbl moduleName) clauses)


data Result
  = ResultUnknown
  | ResultFalse
  | ResultValues !(Map Name (Set Int))

instance Monoid Result where
    mempty = ResultUnknown

    mappend ResultFalse _   = ResultFalse
    mappend _ ResultFalse   = ResultFalse
    mappend (ResultValues l) (ResultValues r) =
        let vals = Map.unionWith Set.intersection l r
        in normalize (ResultValues vals)
    mappend ResultUnknown r = r
    mappend r ResultUnknown = r

normalize :: Result -> Result
normalize (ResultValues vals)
  | any Set.null (Map.elems vals) = ResultFalse
  | otherwise = ResultValues vals
normalize r = r


analyzeClause :: Constants -> SymbolTable -> Name -> Expr () -> Result
analyzeClause consts symTbl moduleName e = case freeVars consts e of
    [] -> case analyzeClosedClause consts e of
        Just False -> ResultFalse
        _          -> ResultUnknown
    [var] -> case Map.lookup (LocalName moduleName var) symTbl of
        Just (Range lower upper) ->
            let vals = filter (couldBeTrue var e) [lower .. upper]
            in ResultValues (Map.singleton var (Set.fromList vals))
        Nothing -> ResultUnknown
    _ -> ResultUnknown
  where
    couldBeTrue var e' i =
        case analyzeClosedClause consts (instantiate var e' i) of
            Just False -> False
            _          -> True


analyzeClosedClause :: Constants -> Expr a -> Maybe Bool
analyzeClosedClause consts = \case
    BoolExpr b _ -> Just b
    UnaryExpr (LogicUnOp LNot) e _ -> not <$> analyzeClosedClause consts e
    BinaryExpr (EqBinOp eqOp) l r _ ->
        let f = case eqOp of
                    Eq  -> (==)
                    Neq -> (/=)
        in f <$> eval consts l <*> eval consts r
    _ -> Nothing


instantiate :: Name -> Expr () -> Int -> Expr ()
instantiate name e i =
    substitute name (IntegerExpr (fromIntegral i) ()) e


freeVars :: Constants -> Expr a -> [Name]
freeVars consts =
    filter (not . (`Map.member` consts)) . mapMaybe getName . universe
  where
    getName = \case
        IdentExpr name _ -> Just name
        _                -> Nothing
