//==============================================================================
//	
//	Copyright (c) 2002-
//	Authors:
//	* Dave Parker <david.parker@comlab.ox.ac.uk> (University of Oxford, formerly University of Birmingham)
//	
//------------------------------------------------------------------------------
//	
//	This file is part of PRISM.
//	
//	PRISM is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	PRISM is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with PRISM; if not, write to the Free Software Foundation,
//	Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//==============================================================================

package parser.ast;

import parser.visitor.*;
import prism.PrismLangException;

public class Command extends ASTElement
{
	/** Synch information */
	private SynchList synchs = new SynchList();
	/** Is this an open or closed action? */
	private boolean open;
	// Guard
	private Expression guard;
	// List of updates
	private Updates updates;
	// Parent module
	private Module parent;
	
	// Constructor
	
	public Command()
	{
		guard = null;
		updates = null;
		open = false;
	}
	
	// Set methods

	public void setSynchs(SynchList list)
	{
		synchs = list;
	}

	/** Set a singleton action */
	public void setSynch(String s) throws PrismLangException
	{
		synchs.clear();
		if (!s.equals("")) {
			synchs.addSynch(s);
		}
	}

	@Deprecated
	public void addSynch(String s) throws PrismLangException
	{
		synchs.addSynch(s);
	}

	public void setOpenSynch(boolean value)
	{
		open = value;
	}

	public void setGuard(Expression g)
	{
		guard = g;
	}
	
	public void setUpdates(Updates u)
	{
		updates = u;
		u.setParent(this);
	}
	
	public void setParent(Module m)
	{
		parent = m;
	}

	// Get methods

	public SynchList getSynchs()
	{
		return synchs;
	}

	/**
	 * Get the action label for this command. For independent ("tau"-labelled) commands,
	 * this is the empty string "" (it should never be null).
	 * <br>
	 * Only applicable in the case of a singleton action, otherwise throws a PrismLangException
	 */
	public String getSynch() throws PrismLangException
	{
		if (synchs.cardinality() > 1) {
			throw new PrismLangException("getSynch called for multi-action");
		}
		if (synchs.cardinality() == 0) {
			// tau action
			return "";
		}
		return synchs.getSynch(0);
	}
	
	/**
	 * Get the index of the action label for this command (in the model's list of actions).
	 * This is 1-indexed, with 0 denoting an independent ("tau"-labelled) command.
	 * -1 denotes not (yet) known.
	 * <br>
	 * Only applicable in the case of a singleton action, otherwise throws a PrismLangException
	 */
	public int getSynchIndex() throws PrismLangException
	{
		if (synchs.cardinality() > 1) {
			throw new PrismLangException("getSynchIndex called for multi-action");
		}
		if (synchs.cardinality() == 0) {
			// tau action = index 0
			return 0;
		}
		return synchs.getSynchIndex(0);
	}

	public boolean getOpenSynch()
	{
		return open;
	}
	
	public boolean hasTauSynch()
	{
		return synchs.isEmpty();
	}

	public Expression getGuard()
	{
		return guard;
	}
	
	public Updates getUpdates()
	{
		return updates;
	}
	
	public Module getParent()
	{
		return parent;
	}
	
	// Methods required for ASTElement:
	
	/**
	 * Visitor method.
	 */
	public Object accept(ASTVisitor v) throws PrismLangException
	{
		return v.visit(this);
	}
	
	/**
	 * Convert to string.
	 */
	public String toString()
	{
		String s ="";
		if (open) {
			s += "]" + synchs + "[ ";
		} else {
			s += "[" + synchs + "] ";
		}
		s += guard + " -> " + updates;
		return s;
	}
	
	/**
	 * Perform a deep copy.
	 */
	public ASTElement deepCopy()
	{
		Command ret = new Command();
		ret.synchs = synchs.clone();
		ret.setOpenSynch(getOpenSynch());
		ret.setGuard(getGuard().deepCopy());
		ret.setUpdates((Updates)getUpdates().deepCopy());
		ret.setPosition(this);
		return ret;
	}

}

//------------------------------------------------------------------------------
