//==============================================================================
//	
//	Copyright (c) 2002-
//	Authors:
//	* Dave Parker <david.parker@comlab.ox.ac.uk> (University of Oxford, formerly University of Birmingham)
//	
//------------------------------------------------------------------------------
//	
//	This file is part of PRISM.
//	
//	PRISM is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	PRISM is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with PRISM; if not, write to the Free Software Foundation,
//	Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//==============================================================================

package parser.ast;

import java.util.ArrayList;

import parser.visitor.*;
import prism.Pair;
import prism.PrismLangException;

public class RenamedModule extends ASTElement
{
	// Rename info
	private String name;
	private String baseModule;
	private ArrayList<Pair<String,String>> rules;
	private ArrayList<String> oldNames;
	private ArrayList<String> newNames;
	// AST elements (for positional info)
	private ExpressionIdent nameASTElement;
	private ExpressionIdent baseModuleASTElement;
	private ArrayList<ExpressionIdent> oldNameASTElements;
	private ArrayList<ExpressionIdent> newNameASTElements;
	
	// Is this a template module, i.e., will only be instantiated via further module renaming?
	private boolean isTemplate;

	// Constructor
	
	public RenamedModule(String n, String b)
	{
		name = n;
		baseModule = b;
		oldNames = new ArrayList<String>();
		newNames = new ArrayList<String>();
		oldNameASTElements = new ArrayList<ExpressionIdent>();
		newNameASTElements = new ArrayList<ExpressionIdent>();
		isTemplate = false;
	}
	
	// Set methods
	
	public void setName(String n)
	{
		name = n;
	}

	public void setBaseModule(String b)
	{
		baseModule = b;
	}

	/** Sets flag whether this module is a template module */
	public void setIsTemplate(boolean value)
	{
		isTemplate = value;
	}

	public void setNameASTElement(ExpressionIdent e)
	{
		nameASTElement = e;
	}
	
	public void setBaseModuleASTElement(ExpressionIdent e)
	{
		baseModuleASTElement = e;
	}
	
	public void addRename(String s1, String s2)
	{
		addRename(s1, s2, null, null);
	}

	public void addRename(String s1, String s2, ExpressionIdent e1, ExpressionIdent e2)
	{
		oldNames.add(s1);
		newNames.add(s2);
		oldNameASTElements.add(e1);
		newNameASTElements.add(e2);
	}

	public void prependRename(String s1, String s2)
	{
		oldNames.add(0, s1);
		newNames.add(0, s2);
		oldNameASTElements.add(null);
		newNameASTElements.add(null);
	}

	public void addRule(String key, String value)
	{
		if (rules == null)
			rules = new ArrayList<>();
		rules.add(new Pair<>(key,value));
	}
		
	// Get methods
	
	public String getName()
	{
		return name;
	}
	
	public String getBaseModule()
	{
		return baseModule;
	}

	/** Returns true if this is a template module */
	public boolean isTemplate()
	{
		return isTemplate;
	}

	public String getNewName(String s)
	{
		int i = oldNames.indexOf(s);
		if (i == -1) {
			return null;
		}
		return newNames.get(i);
	}
		
	public String getOldName(String s)
	{
		int i = newNames.indexOf(s);
		if (i == -1) {
			return null;
		}
		return oldNames.get(i);
	}
	
	public int getNumRenames()
	{
		return oldNames.size();
	}
	
	public String getOldName(int i)
	{
		return oldNames.get(i); 
	}
	
	public String getNewName(int i)
	{
		return newNames.get(i); 
	}
	
	public ExpressionIdent getNameASTElement()
	{
		return nameASTElement;
	}
	
	public ExpressionIdent getBaseModuleASTElement()
	{
		return baseModuleASTElement;
	}
	
	public ExpressionIdent getOldNameASTElement(int i)
	{
		return oldNameASTElements.get(i);
	}
	
	public ExpressionIdent getNewNameASTElement(int i)
	{
		return newNameASTElements.get(i);
	}
	
	public boolean hasRules()
	{
		return rules != null && !rules.isEmpty();
	}
	
	public void expandRules(Module baseModule) throws PrismLangException
	{
		if (!hasRules())
			return;

		String varPrefix = "";
		String varSuffix = "";
		String actionPrefix = "";
		String actionSuffix = "";

		for (Pair<String,String> r : rules) {
			switch (r.first) {
			case "varprefix":
				varPrefix = r.second + varPrefix;
				break;
			case "varsuffix":
				varSuffix += r.second;
				break;
			case "actionprefix":
				actionPrefix = r.second + actionPrefix;
				break;
			case "actionsuffix":
				actionSuffix += r.second;
				break;
			default:
				throw new PrismLangException("Unknown rule '" + r.first + "' in renaming", this);
			}
		}

		if (!varPrefix.isEmpty() || !varSuffix.isEmpty()) {
			for (int i = 0, n = baseModule.getNumDeclarations(); i < n; i++) {
				String var = baseModule.getDeclaration(i).getName();
				String renamedVar = varPrefix + var + varSuffix;
				prependRename(var, renamedVar);
			}
		}

		if (!actionPrefix.isEmpty() || !actionSuffix.isEmpty()) {
			for (String synch : baseModule.getAllSynchs()) {
				String renamedSynch = actionPrefix + synch + actionSuffix;
				prependRename(synch, renamedSynch);
			}
		}

	}

	// Methods required for ASTElement:
	
	/**
	 * Visitor method.
	 */
	public Object accept(ASTVisitor v) throws PrismLangException
	{
		return v.visit(this);
	}
	
	/**
	 * Convert to string.
	 */
	public String toString()
	{
		String s = "";
		int i;
		
		s = s + "module " + name + " = " + baseModule;
		if (rules != null) {
			s += "(";
			boolean first = true;
			for (Pair<String, String> r : rules) {
				if (!first) s+= ",";
				first = false;
				s += r.first + "," + r.second;
			}
			s += ")";
		}
		s += " [";
		boolean first = true;
		for (i = 0; i < oldNames.size(); i++) {
			if (!first) s+= ", ";
			first = false;
			s = s + oldNames.get(i) + " = " + newNames.get(i);
		}
		s += "] " + (isTemplate ? "as template " : "" ) + "endmodule";
		
		return s;
	}
	
	/**
	 * Perform a deep copy.
	 */
	public ASTElement deepCopy()
	{
		int i, n;
		RenamedModule ret = new RenamedModule(name, baseModule);
		ret.isTemplate = isTemplate;
		ret.setNameASTElement((ExpressionIdent)nameASTElement.deepCopy());
		ret.setBaseModuleASTElement((ExpressionIdent)baseModuleASTElement.deepCopy());
		n = oldNames.size();
		for (i = 0; i < n; i++) {
			ret.addRename(oldNames.get(i), newNames.get(i));
		}
		if (hasRules()) {
			for (Pair<String, String> r : rules) {
				ret.addRule(r.first, r.second);
			}
		}
		ret.setPosition(this);
		return ret;
	}
}

//------------------------------------------------------------------------------
