package parser.ast;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import prism.PrismLangException;

public class SynchList implements Iterable<String>
{
	/** List of synchs */
	private List<String> synchs = new ArrayList<String>();

	/**
	 * Indices of the elements of the synchs list in model's list of all actions ("synchs")
	 * This is 1-indexed, with 0 denoting an independent ("tau"-labelled) command.
	 * -1 denotes not (yet) known.
	 */
	private List<Integer> synchIndizes = new ArrayList<Integer>();

	public void addSynch(String s)
	{
		addSynch(s, -1);
	}

	public void addSynch(String s, int index)
	{
		if (synchs.contains(s)) {
			// ignore duplicate
			return;
		}
		if (s.equals("")) {
			throw new IllegalArgumentException("Can not add tau-action to multi-action");
		}
		synchs.add(s);
		synchIndizes.add(index);
	}

	public void setSynch(int i, String s)
	{
		setSynch(i, s, -1);
	}
	
	public void setSynch(int i, String s, int sIndex)
	{
		synchs.set(i, s);
		synchIndizes.set(i, sIndex);
	}

	public void setSynchIndex(int i, int sIndex)
	{
		synchIndizes.set(i, sIndex);
	}

	public void clear()
	{
		synchs.clear();
		synchIndizes.clear();
	}

	public boolean isEmpty()
	{
		return synchs.isEmpty();
	}

	public int cardinality()
	{
		return synchs.size();
	}

	public String getSynch(int i)
	{
		return synchs.get(i);
	}

	public int getSynchIndex(int i)
	{
		return synchIndizes.get(i);
	}

	public List<String> getSynchs()
	{
		return Collections.unmodifiableList(synchs);
	}

	public List<Integer> getSynchIndizes()
	{
		return Collections.unmodifiableList(synchIndizes);
	}

	public BitSet asBitSet()
	{
		BitSet bs = new BitSet();
		for (int i : synchIndizes) {
			bs.set(i);
		}
		return bs;
	}

	public Iterator<String> iterator()
	{
		return Collections.unmodifiableList(synchs).iterator();
	}

	@Override
	public String toString()
	{
		if (synchs.isEmpty()) {
			return "";
		} else if (synchs.size() == 1) {
			return synchs.get(0);
		} else {
			StringBuilder s = new StringBuilder();
			boolean first = true;
			for (String synch : synchs) {
				if (!first) s.append(",");
				first = false;
				s.append(synch);
			}
			return s.toString();
		}
	}

	public SynchList clone()
	{
		SynchList result = new SynchList();

		result.synchs.addAll(synchs);
		result.synchIndizes.addAll(synchIndizes);

		return result;
	}

}
