//==============================================================================
//	
//	Copyright (c) 2002-
//	Authors:
//	* Dave Parker <d.a.parker@cs.bham.ac.uk> (University of Birmingham/Oxford)
//	
//------------------------------------------------------------------------------
//	
//	This file is part of PRISM.
//	
//	PRISM is free software; you can redistribute it and/or modify
//	it under the terms of the GNU General Public License as published by
//	the Free Software Foundation; either version 2 of the License, or
//	(at your option) any later version.
//	
//	PRISM is distributed in the hope that it will be useful,
//	but WITHOUT ANY WARRANTY; without even the implied warranty of
//	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//	GNU General Public License for more details.
//	
//	You should have received a copy of the GNU General Public License
//	along with PRISM; if not, write to the Free Software Foundation,
//	Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
//	
//==============================================================================

package simulator;

import java.util.ArrayList;
import java.util.BitSet;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.PrimitiveIterator.OfInt;
import java.util.TreeMap;
import java.util.Vector;

import common.IterableBitSet;
import parser.State;
import parser.VarList;
import parser.ast.Command;
import parser.ast.Module;
import parser.ast.ModulesFile;
import parser.ast.RewardStruct;
import parser.ast.RewardStructItem;
import parser.ast.Update;
import parser.ast.Updates;
import prism.ModelType;
import prism.Pair;
import prism.PrismComponent;
import prism.PrismException;
import prism.PrismLangException;
import prism.PrismSettings;

public class Updater extends PrismComponent
{
	// Settings:
	// Do we check that probabilities sum to 1?
	protected boolean doProbChecks = true;
	// The precision to which we check probabilities sum to 1
	protected double sumRoundOff = 1e-5;
	
	// Info on model being explored
	protected ModulesFile modulesFile;
	protected ModelType modelType;
	protected int numModules;
	protected VarList varList;
	// Synchronising action info
	protected ArrayList<String> synchs;
	protected int numSynchs;
	protected int synchModuleCounts[];
	protected BitSet synchAlphabetPerModule[];
	// Model info/stats
	protected int numRewardStructs;

	private static boolean debug = false;
	
	// Temporary storage:

	private static class EnabledCommands {
		public Map<BitSet, List<Updates>> map = new LinkedHashMap<BitSet, List<Updates>>();

		public List<Updates> forAction(BitSet action) {
			List<Updates> u = map.get(action);
			if (u == null) {
				return Collections.emptyList();
			}
			return u;
		}

		public void addUpdatesForAction(BitSet action, Updates u)
		{
			List<Updates> updates = map.get(action);
			if (updates == null) {
				updates = new ArrayList<Updates>();
				map.put(action, updates);
			}
			updates.add(u);
		}
		
		public void addAll(BitSet action, Collection<Updates> ups) {
			List<Updates> updates = map.get(action);
			if (updates == null) {
				updates = new ArrayList<Updates>();
				map.put(action, updates);
			}
			updates.addAll(ups);
		}

		public Iterable<BitSet> getActions()
		{
			return map.keySet();
		}

		public String toString()
		{
			return map.toString();
		}

		public void clear()
		{
			map.clear();
		}
	}

	private class EnabledChoices {
		public Map<BitSet, List<ChoiceListFlexi>> map = new LinkedHashMap<BitSet, List<ChoiceListFlexi>>();

		public List<ChoiceListFlexi> forAction(BitSet action)
		{
			return map.get(action);
		}

		public void add(BitSet action, List<ChoiceListFlexi> chs)
		{
			List<ChoiceListFlexi> list = forAction(action);
			if (list == null) {
				list = new ArrayList<ChoiceListFlexi>();
				map.put(action, list);
			}
			list.addAll(chs);
		}

		public Iterable<BitSet> getActions()
		{
			return map.keySet();
		}

		public void clear()
		{
			map.clear();
		}

		public String toString()
		{
			return map.toString();
		}

	}

	
	// Temporary storage:

	// Element i of enabledPerModule is a list of the updates from module i labelled with action j
	// (where j=0 denotes independent, otherwise 1-indexed action label)
	protected ArrayList<EnabledCommands> enabledPerModuleClosed;
	protected ArrayList<EnabledCommands> enabledPerModuleOpen;

	protected EnabledChoices enabledOpen = new EnabledChoices();
	protected EnabledChoices enabledClosed = new EnabledChoices();
	protected BitSet alphabet;

	public Updater(ModulesFile modulesFile, VarList varList)
	{
		this(modulesFile, varList, null);
	}
	
	public Updater(ModulesFile modulesFile, VarList varList, PrismComponent parent)
	{
		// Store some settings
		doProbChecks = parent.getSettings().getBoolean(PrismSettings.PRISM_DO_PROB_CHECKS);
		sumRoundOff = parent.getSettings().getDouble(PrismSettings.PRISM_SUM_ROUND_OFF);
		
		// Get info from model
		this.modulesFile = modulesFile;
		modelType = modulesFile.getModelType();
		numModules = modulesFile.getNumModules();
		synchs = new ArrayList<>(modulesFile.getSynchs());
		numSynchs = synchs.size();
		numRewardStructs = modulesFile.getNumRewardStructs();
		this.varList = varList;

		// Compute count of number of modules using each synch action
		// First, compute and cache the synch actions for each of the modules
		List<HashSet<String>> synchsPerModule = new ArrayList<HashSet<String>>(numModules);
		synchAlphabetPerModule = new BitSet[numModules];
		for (int i = 0; i < numModules; i++) {
			synchsPerModule.add(new HashSet<String>(modulesFile.getModule(i).getAllSynchs()));
			synchAlphabetPerModule[i] = new BitSet();
		}
		// Second, do the counting
		synchModuleCounts = new int[numSynchs];
		for (int j = 0; j < numSynchs; j++) {
			synchModuleCounts[j] = 0;
			String s = synchs.get(j);
			for (int i = 0; i < numModules; i++) {
				if (synchsPerModule.get(i).contains(s)) {
					synchModuleCounts[j]++;
					synchAlphabetPerModule[i].set(j+1);  // indizes are starting from 1
				}
			}
		}

		// Build temporary storage for later use
		enabledPerModuleClosed = new ArrayList<EnabledCommands>(numModules);
		enabledPerModuleOpen = new ArrayList<EnabledCommands>(numModules);
		for (int i = 0; i < numModules; i++) {
			enabledPerModuleClosed.add(new EnabledCommands());
			enabledPerModuleOpen.add(new EnabledCommands());
		}
		alphabet = new BitSet();
	}

	/**
	 * Set the precision to which we check that probabilities sum to 1.
	 */
	public void setSumRoundOff(double sumRoundOff)
	{
		this.sumRoundOff = sumRoundOff;
	}

	/**
	 * Get the precision to which we check that probabilities sum to 1.
	 */
	public double getSumRoundOff()
	{
		return sumRoundOff;
	}

	/**
	 * Determine the set of outgoing transitions from state 'state' and store in 'transitionList'.
	 * @param state State from which to explore
	 * @param transitionList TransitionList object in which to store result
	 */
	public void calculateTransitions(State state, TransitionList transitionList) throws PrismException
	{
		int i, j, k, l, n, count;

		// Clear lists/bitsets
		transitionList.clear();
		for (i = 0; i < numModules; i++) {
			enabledPerModuleClosed.get(i).clear();
			enabledPerModuleOpen.get(i).clear();
		}
		enabledClosed.clear();
		enabledOpen.clear();
		alphabet.clear();

		// Calculate the available updates for each module/action
		// (update information in updateLists, enabledSynchs and enabledModules)
		for (i = 0; i < numModules; i++) {
			calculateUpdatesForModule(i, state);
		}
		//System.out.println("updateLists: " + updateLists);

		/*
		// Add independent transitions (closed) for each (enabled) module to list
		BitSet tauAction = new BitSet();
		for (i = 0; i < numModules; i++) {
			for (Updates ups : enabledPerModuleClosed.get(i).forAction(tauAction)) {
				ChoiceListFlexi ch = processUpdatesAndCreateNewChoice(-(i + 1), ups, state);
				if (ch.size() > 0)
					transitionList.add(ch);
			}
		}
		 */

		if (debug) {
			for (i = 0; i < numModules; i++) {
				System.out.println("Module " + i + " [] = " + enabledPerModuleClosed.get(i));
				System.out.println("Module " + i + " ][ = " + enabledPerModuleOpen.get(i));
			}
		}

		BitSet tauAction = new BitSet();

		for (i = 0; i < numModules; i++) {
			if (i == 0) {
				for (BitSet alpha : enabledPerModuleClosed.get(0).getActions()) {
					enabledClosed.add(alpha, create(state, enabledPerModuleClosed.get(0).forAction(alpha)));
				}
				for (BitSet alpha : enabledPerModuleOpen.get(0).getActions()) {
					enabledOpen.add(alpha, create(state, enabledPerModuleOpen.get(0).forAction(alpha)));
				}

				alphabet.or(synchAlphabetPerModule[0]);
				if (debug) {
					System.out.println("------------ after " + i + " ---------------------");
					System.out.println("[]: " + enabledClosed);
					System.out.println("][: " + enabledOpen);
					System.out.println("Alphabet: " + alphabet);
				}
				continue;
			}

			BitSet A = alphabet;
			BitSet B = synchAlphabetPerModule[i];
			BitSet AcapB = (BitSet) A.clone();
			AcapB.and(B);

			EnabledChoices newEnabledOpen = new EnabledChoices();
			EnabledChoices newEnabledClosed = new EnabledChoices();

			EnabledCommands enabledBOpen = enabledPerModuleOpen.get(i);
			EnabledCommands enabledBClosed = enabledPerModuleClosed.get(i);
			
			for (BitSet alpha : enabledClosed.getActions()) {
				if (alpha.isEmpty()) {
					// (?) keep, but don't sync
					newEnabledClosed.add(alpha, enabledClosed.forAction(alpha));
					continue;
				}

				// (4) [y] := [a] for [a] || ]b[ with a \cap B = b
				BitSet beta = (BitSet) alpha.clone();
				beta.and(B);
				List<Updates> updateList = enabledBOpen.forAction(beta);
				if (!updateList.isEmpty()) {
					newEnabledClosed.add(alpha, product(state, enabledClosed.forAction(alpha), updateList));
				}

				if (!alpha.intersects(B)) {
					// (1) [y] := [a]
					newEnabledClosed.add(alpha, enabledClosed.forAction(alpha));
				} else {
					// (3) [y] := [a] = [b]
					// fetch corresponding [b]
					updateList = enabledBClosed.forAction(alpha);
					if (!updateList.isEmpty()) {
						newEnabledClosed.add(alpha, product(state, enabledClosed.forAction(alpha), updateList));
					}
				}
			}

			for (BitSet beta : enabledBClosed.getActions()) {
				if (beta.isEmpty()) {
					// (?) [b] add, but don't sync
					newEnabledClosed.add(beta, create(state, enabledBClosed.forAction(beta)));
					continue;
				}

				if (!beta.intersects(A)) {
					// (2) [y] := [b]
					newEnabledClosed.add(beta, create(state, enabledBClosed.forAction(beta)));
				}
				// ignore, case (3), handled above from the other side

				// (5) [y] := [b] for [b] || ]a[ with b \cap A = a
				BitSet alpha = (BitSet) beta.clone();
				alpha.and(A);
				if (enabledOpen.forAction(alpha) != null) {
					// produce [b] from [b] || ]a[ 
					newEnabledClosed.add(beta, product(state, enabledOpen.forAction(alpha), enabledBClosed.forAction(beta)));
				}
			}

			HashMap<BitSet, List<Pair<BitSet,List<ChoiceListFlexi>>>> openA = new LinkedHashMap<>();
			// compute shared actions
			for (BitSet alpha : enabledOpen.getActions()) {
				BitSet alphaB = (BitSet) alpha.clone();
				alphaB.and(B);
				List<Pair<BitSet, List<ChoiceListFlexi>>> list = openA.get(alphaB);
				if (list == null) {
					list = new ArrayList<>();
					openA.put(alphaB, list);
				}
				list.add(new Pair<>(alpha, enabledOpen.forAction(alpha)));
			}
			HashMap<BitSet, List<Pair<BitSet,List<Updates>>>> openB = new HashMap<>();
			for (BitSet beta : enabledBOpen.getActions()) {
				BitSet betaA = (BitSet) beta.clone();
				betaA.and(A);
				List<Pair<BitSet, List<Updates>>> list = openB.get(betaA);
				if (list == null) {
					list = new ArrayList<>();
					openB.put(betaA, list);
				}
				list.add(new Pair<>(beta, enabledBOpen.forAction(beta)));
			}

			if (openA.get(tauAction) != null) {
				// (6) ]y[ := ]a[ for a with a cap B = 0
				for (Pair<BitSet, List<ChoiceListFlexi>> act : openA.get(tauAction)) {
					newEnabledOpen.add(act.first, act.second);
				}
				
			}
			if (openB.get(tauAction) != null) {
				// (6) ]y[ := ]b[ for b with b cap B = 0
				for (Pair<BitSet, List<Updates>> act : openB.get(tauAction)) {
					newEnabledOpen.add(act.first, create(state, act.second));
				}
			}

			for (BitSet alpha : enabledOpen.getActions()) {
				BitSet alphaB = (BitSet) alpha.clone();
				alphaB.and(B);

				List<Pair<BitSet, List<Updates>>> matchB = openB.get(alphaB);
				if (matchB == null) {
					continue;
				}
				for (Pair<BitSet, List<Updates>> act : matchB) {
					BitSet alpha_beta = (BitSet) alpha.clone();
					alpha_beta.or(act.first);  // alpha_beta = alpha cup beta
					newEnabledOpen.add(alpha_beta, product(state, enabledOpen.forAction(alpha), act.second));
				}
			}

			alphabet.or(synchAlphabetPerModule[i]);
			enabledOpen = newEnabledOpen;
			enabledClosed = newEnabledClosed;

			if (debug) {
				System.out.println("------------ after " + i + " ---------------------");
				System.out.println("[]: " + enabledClosed);
				System.out.println("][: " + enabledOpen);
				System.out.println("Alphabet: " + alphabet);
			}
		}

		if (debug) {
			System.out.println("------------ Final ---------------------");
			System.out.println("[]: " + enabledClosed);
			System.out.println("][: " + enabledOpen);
			System.out.println("Alphabet: " + alphabet);
		}


		// Add all choices to transition list
		for (Entry<BitSet, List<ChoiceListFlexi>> entry : enabledClosed.map.entrySet()) {
			BitSet action = entry.getKey();
			MultiAction multiAction = new MultiAction(action, synchs);
			List<ChoiceListFlexi> chs = entry.getValue();
			for (ChoiceListFlexi ch : chs) {
				ch.setMultiAction(multiAction);
				transitionList.add(ch);
			}
		}
		for (Entry<BitSet, List<ChoiceListFlexi>> entry : enabledOpen.map.entrySet()) {
			BitSet action = entry.getKey();
			MultiAction multiAction = new MultiAction(action, synchs);
			List<ChoiceListFlexi> chs = entry.getValue();
			for (ChoiceListFlexi ch : chs) {
				ch.setMultiAction(multiAction);
				transitionList.add(ch);
			}
		}
		
		// For a DTMC, we need to normalise across all transitions
		// This is partly to handle "local nondeterminism"
		// and also to handle any dubious trickery done by disabling probability checks
		if (modelType == ModelType.DTMC) {
			double probSum = transitionList.getProbabilitySum();
			transitionList.scaleProbabilitiesBy(1.0 / probSum);
		}
	
		// Check validity of the computed transitions
		// (not needed currently)
		//transitionList.checkValid(modelType);
		
		// Check for errors (e.g. overflows) in the computed transitions
		//transitionList.checkForErrors(state, varList);
		
		//System.out.println(transitionList);
	}

	/**
	 * Calculate the state rewards for a given state.
	 * @param state The state to compute rewards for
	 * @param store An array in which to store the rewards
	 */
	public void calculateStateRewards(State state, double[] store) throws PrismLangException
	{
		int i, j, n;
		double d;
		RewardStruct rw;
		for (i = 0; i < numRewardStructs; i++) {
			rw = modulesFile.getRewardStruct(i);
			n = rw.getNumItems();
			d = 0.0;
			for (j = 0; j < n; j++) {
				if (!rw.getRewardStructItem(j).isTransitionReward())
					if (rw.getStates(j).evaluateBoolean(state))
						d += rw.getReward(j).evaluateDouble(state);
			}
			store[i] = d;
		}
	}

	/**
	 * Calculate the transition rewards for a given state and outgoing choice.
	 * @param state The state to compute rewards for
	 * @param ch The choice from the state to compute rewards for
	 * @param store An array in which to store the rewards
	 */
	public void calculateTransitionRewards(State state, Choice ch, double[] store) throws PrismLangException
	{
		int i, j, n;
		double d;
		RewardStruct rw;
		for (i = 0; i < numRewardStructs; i++) {
			rw = modulesFile.getRewardStruct(i);
			n = rw.getNumItems();
			d = 0.0;
			for (j = 0; j < n; j++) {
				RewardStructItem rsi = rw.getRewardStructItem(j);
				if (rsi.isTransitionReward())
					if (rsi.matches(ch.getMultiAction()))
						if (rw.getStates(j).evaluateBoolean(state))
							d += rw.getReward(j).evaluateDouble(state);
			}
			store[i] = d;
		}
	}
	
	// Private helpers
	
	/**
	 * Determine the enabled updates for the 'm'th module from (global) state 'state'.
	 * Update information in updateLists, enabledSynchs and enabledModules.
	 * @param m The module index
	 * @param state State from which to explore
	 */
	protected void calculateUpdatesForModule(int m, State state) throws PrismLangException
	{
		Module module;
		Command command;
		int i, n;

		module = modulesFile.getModule(m);
		EnabledCommands enabledOpen = enabledPerModuleOpen.get(m);
		EnabledCommands enabledClosed = enabledPerModuleClosed.get(m);
		n = module.getNumCommands();
		for (i = 0; i < n; i++) {
			command = module.getCommand(i);
			if (command.getGuard().evaluateBoolean(state)) {
				BitSet action = command.getSynchs().asBitSet();
				if (command.getOpenSynch()) {
					enabledOpen.addUpdatesForAction(action, command.getUpdates());
				} else {
					enabledClosed.addUpdatesForAction(action, command.getUpdates());
				}
			}
		}

		if (debug) {
			System.out.println("\nModule " + module.getName() + " open for state " + state + ": " + enabledOpen);
			System.out.println("Module " + module.getName() + " closed for state " + state + ": " + enabledClosed);
		}
	}

	/**
	 * Create a new Choice object (currently ChoiceListFlexi) based on an Updates object
	 * and a (global) state. Check for negative probabilities/rates and, if appropriate,
	 * check probabilities sum to 1 too.
	 * @param moduleOrActionIndex Module/action for the choice, encoded as an integer (see Choice)
	 * @param ups The Updates object 
	 * @param state Global state
	 */
	private ChoiceListFlexi processUpdatesAndCreateNewChoice(int moduleOrActionIndex, Updates ups, State state) throws PrismLangException
	{
		ChoiceListFlexi ch;
		List<Update> list;
		int i, n;
		double p, sum;

		// Create choice and add all info
		ch = new ChoiceListFlexi();
		ch.setModuleOrActionIndex(moduleOrActionIndex);
		n = ups.getNumUpdates();
		sum = 0;
		for (i = 0; i < n; i++) {
			// Compute probability/rate
			p = ups.getProbabilityInState(i, state);
			// Check for negative/NaN probabilities/rates
			if (Double.isNaN(p) || p < 0) {
				String s = modelType.choicesSumToOne() ? "Probability" : "Rate";
				s += " is invalid (" + p + ") in state " + state.toString(modulesFile);
				// Note: we indicate error in whole Updates object because the offending
				// probability expression has probably been simplified from original form.
				throw new PrismLangException(s, ups);
			}
			// Skip transitions with zero probability/rate
			if (p == 0)
				continue;
			sum += p;
			list = new ArrayList<Update>();
			list.add(ups.getUpdate(i));
			ch.add(p, list);
		}
		// For now, PRISM treats empty (all zero probs/rates) distributions as an error.
		// Later, when errors in symbolic model construction are improved, this might be relaxed.
		if (ch.size() == 0) {
			String msg = modelType.probabilityOrRate();
			msg += (ups.getNumUpdates() > 1) ? " values sum to " : " is ";
			msg += "zero for updates in state " + state.toString(modulesFile);
			throw new PrismLangException(msg, ups);
		}
		// Check distribution sums to 1 (if required, and if is non-empty)
		if (doProbChecks && ch.size() > 0 && modelType.choicesSumToOne() && Math.abs(sum - 1) > sumRoundOff) {
			throw new PrismLangException("Probabilities sum to " + sum + " in state " + state.toString(modulesFile), ups);
		}
		return ch;
	}

	/**
	 * Create a new Choice object (currently ChoiceListFlexi) based on the product
	 * of an existing ChoiceListFlexi and an Updates object, for some (global) state.
	 * If appropriate, check probabilities sum to 1 too.
	 * @param ups The Updates object 
	 * @param state Global state
	 * @param ch The existing Choices object
	 */
	private ChoiceListFlexi processUpdatesAndGenerateProduct(Updates ups, State state, ChoiceListFlexi ch) throws PrismLangException
	{
		ChoiceListFlexi chProduct = new ChoiceListFlexi(ch);
		// Create new choice (action index is 0 - not needed)
		ChoiceListFlexi chNew = processUpdatesAndCreateNewChoice(0, ups, state);
		// Build product with existing
		chProduct.productWith(chNew);
		return chProduct;
	}

	
	/**
	 * Create a new Choice object (currently ChoiceListFlexi) based on the product
	 * of an existing ChoiceListFlexi and an Updates object, for some (global) state.
	 * If appropriate, check probabilities sum to 1 too.
	 * @param ups The Updates object 
	 * @param state Global state
	 * @param ch The existing Choices object
	 */
	private void processUpdatesAndAddToProduct(Updates ups, State state, ChoiceListFlexi ch) throws PrismLangException
	{
		// Create new choice (action index is 0 - not needed)
		ChoiceListFlexi chNew = processUpdatesAndCreateNewChoice(0, ups, state);
		// Build product with existing
		ch.productWith(chNew);
	}

	private List<ChoiceListFlexi> create(State state, List<Updates> updateList) throws PrismLangException
	{
		ArrayList<ChoiceListFlexi> result = new ArrayList<ChoiceListFlexi>();
		for (Updates ups: updateList) {
			result.add(processUpdatesAndCreateNewChoice(0, ups, state));
		}
		return result;
	}
	
	private List<ChoiceListFlexi> product(State state, List<ChoiceListFlexi> chs, List<Updates> updateList) throws PrismLangException
	{
		List<ChoiceListFlexi> result = new ArrayList<ChoiceListFlexi>();
		
		for (Updates ups : updateList) {
			// Product with all existing choices
			for (ChoiceListFlexi ch : chs) {
				result.add(processUpdatesAndGenerateProduct(ups, state, ch));
			}
		}
		return result;
	}
	
	private String actionToName(BitSet action)
	{
		StringBuilder sb = new StringBuilder();
		OfInt it = IterableBitSet.getSetBits(action).iterator();
		boolean first = true;
		while (it.hasNext()) {
			if (!first) sb.append(",");
			first = false;
			int i = it.nextInt();
			sb.append(synchs.get(i-1));
		}
		return sb.toString();
	}

}
