package nl.cwi.reo;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.util.*;
import java.util.Map.Entry;
import java.util.function.Function;

import org.stringtemplate.v4.ST;
import org.stringtemplate.v4.STGroup;
import org.stringtemplate.v4.STGroupFile;

import com.beust.jcommander.JCommander;
import com.beust.jcommander.Parameter;

import nl.cwi.reo.commands.Command;
import nl.cwi.reo.commands.Commands;
import nl.cwi.reo.compile.CompilerType;
import nl.cwi.reo.compile.LykosCompiler;
import nl.cwi.reo.interpret.Atom;
import nl.cwi.reo.interpret.ReoProgram;
import nl.cwi.reo.interpret.SemanticsType;
import nl.cwi.reo.interpret.connectors.Language;
import nl.cwi.reo.interpret.connectors.Links;
import nl.cwi.reo.interpret.connectors.Reference;
import nl.cwi.reo.interpret.connectors.ReoConnector;
import nl.cwi.reo.interpret.connectors.ReoConnectorAtom;
import nl.cwi.reo.interpret.connectors.ReoConnectorComposite;
import nl.cwi.reo.interpret.interpreters.Interpreter;
import nl.cwi.reo.interpret.ports.Port;
import nl.cwi.reo.interpret.ports.PortType;
import nl.cwi.reo.interpret.typetags.TypeTag;
import nl.cwi.reo.interpret.values.BooleanValue;
import nl.cwi.reo.interpret.values.DecimalValue;
import nl.cwi.reo.interpret.values.StringValue;
import nl.cwi.reo.interpret.values.Value;
import nl.cwi.reo.pr.comp.CompilerSettings;
import nl.cwi.reo.semantics.Semantics;
import nl.cwi.reo.semantics.hypergraphs.ConstraintHypergraph;
import nl.cwi.reo.semantics.prautomata.ListenerPR;
import nl.cwi.reo.semantics.prba.ListenerPRBA;
import nl.cwi.reo.semantics.prba.RewardMonitor;
import nl.cwi.reo.semantics.predicates.MemoryVariable;
import nl.cwi.reo.semantics.predicates.Term;
import nl.cwi.reo.semantics.rulebasedautomata.ListenerRBA;
import nl.cwi.reo.semantics.rulebasedautomata.Rule;
import nl.cwi.reo.semantics.rulebasedautomata.RuleBasedAutomaton;
import nl.cwi.reo.templates.Atomic;
import nl.cwi.reo.templates.Component;
import nl.cwi.reo.templates.MaudeProtocol;
import nl.cwi.reo.templates.Protocol;
import nl.cwi.reo.templates.ReoTemplate;
import nl.cwi.reo.templates.RewardsAtomic;
import nl.cwi.reo.templates.Transition;
import nl.cwi.reo.util.Message;
import nl.cwi.reo.util.MessageType;
import nl.cwi.reo.util.Monitor;

/**
 * A compiler for the coordination language Reo.
 */
public class Compiler {

	/** Version number. */
	private static String version = "1.0.1";

	/**
	 * List of provided Reo source files.
	 */
	@Parameter(description = ".treo files")
	private List<String> files = new ArrayList<String>();

	/**
	 * Compiler type.
	 */
	@Parameter(names = { "-c" }, description = "compiler")
	public CompilerType compilertype = CompilerType.DEFAULT;

	/** List of of directories that contain all necessary Reo components. */
	@Parameter(names = {
			"-cp" }, variableArity = true, description = "list of directories that contain all necessary Reo components")
	private List<String> directories = new ArrayList<String>();

	/**
	 * List of provided Reo source files.
	 */
	@Parameter(names = { "-o" }, description = "output directory")
	private String outdir = ".";

	/**
	 * List of parameters for the main component.
	 */
	@Parameter(names = {
			"-p" }, variableArity = true, description = "list of parameters to instantiate the main component")
	public List<String> params = new ArrayList<String>();

	/** Package. */
	@Parameter(names = { "-pkg" }, description = "target code package")
	private String packagename;

	/** Partitioning. */
	@Parameter(names = { "-pt" }, description = "synchronous region decomposition")
	private boolean partitioning = false;

	/** Scheduling. */
	@Parameter(names = { "-sch" }, description = "generate custom scheduling policy")
	private boolean scheduling = false;
	
	/** Compose the protocol? */
	@Parameter(names = { "-dontcompose" }, description = "don't compose the protocol circuit")
	private boolean dontComposeProtocol = false;

	/** Protocol connectors first? */
	@Parameter(names = { "-protocolfirst" }, description = "generate protocol connectors first")
	private boolean protocolFirst = false;

	/** Debug output? */
	@Parameter(names = { "-debug" }, description = "debug output")
	private boolean debug = false;

	/**
	 * Target language.
	 */
	@Parameter(names = { "-t" }, variableArity = true, description = "target language")
	public Language lang = Language.JAVA;

	/**
	 * Target language specific include files.
	 */
	@Parameter(names = { "-I" }, variableArity = true, description = "target language specific include file(s)")
	public List<String> includes = new ArrayList<>();

	/**
	 * Language specific settings (either -set key or -set key=value),
	 * as a list.
	 */
	@Parameter(names = { "-set" }, variableArity = true, description = "language specific settings (use -set ? for usage)")
	public List<String> settingArgs = new ArrayList<>();

	/**
	 * Language specific settings, stored as a map key,value, where value is either a string
	 * (for -set key=value) or the Boolean true (for -set key).
	 */
	public Map<String, Object> settings = new TreeMap<>();

	/**
	 * Message container.
	 */
	private final Monitor monitor = new Monitor();

	@SuppressWarnings("serial")
	private static class InvalidArgumentException extends RuntimeException {
		public InvalidArgumentException(String msg) {
			super(msg);
		}
	}
	
	/**
	 * The main method.
	 *
	 * @param args
	 *            the arguments
	 */
	public static void main(String[] args) {
		Compiler compiler = new Compiler();
		try {
			JCommander jc = new JCommander(compiler, args);
			jc.setProgramName("reo");
			if (compiler.settingArgs.contains("?")) {
				String usage = compiler.lang.getSettingsUsage();
				if (usage == null) {
					System.out.println("No language specific parameters for language " + compiler.lang + ":");
				} else {
					System.out.println("Language specific parameters for language " + compiler.lang + ":");
					System.out.println(usage);
				}
			} else if (compiler.files.size() == 0) {
				System.out.println("Reo compiler v" + version + "\nDeveloped at CWI, Amsterdam\n");
				jc.usage();
			} else {
				compiler.run();
			}
		} catch (Exception e) {
			e.printStackTrace();
			System.out.println(new Message(MessageType.ERROR, e.toString()));
		}
	}

	/**
	 * Parse the settings given via the -set option.
	 */
	private void parseSettingsArguments() {
		for (String s : settingArgs) {
			s = s.trim();
			if (s.isEmpty()) {
				throw new InvalidArgumentException("Invalid (empty) argument with -set");
			}
			// find first =
			int posEq = s.indexOf('=');
			String key, value;
			if (posEq == -1) {
				key = s;
				value = null;
			} else {
				key = s.substring(0, posEq);
				value = s.substring(posEq + 1);
				if (key.isEmpty()) {
					throw new InvalidArgumentException("Invalid (empty) key used with -set: " + s);
				}
				if (value.isEmpty()) {
					throw new InvalidArgumentException("Invalid (empty) value used with -set: " + s);
				}
			}
			if (value == null) {
				if (debug) {
					System.out.println(new Message(MessageType.INFO, "Setting flag " + key + " to true"));
				}
				settings.put(key, true);
			} else {
				if (debug) {
					System.out.println(new Message(MessageType.INFO, "Setting " + key + " = " + value));
				}
				settings.put(key, value);
			}
		}
	}

	/**
	 * Run.
	 */
	public void run() {

		directories.add(".");
		String comppath = System.getenv("COMPATH");
		if (comppath != null)
			directories.addAll(Arrays.asList(comppath.split(File.pathSeparator)));

		parseSettingsArguments();
		lang.setDefaults(settings);
		lang.validateSettings(settings);
		for (int i = 0; i < includes.size(); i++) {
			includes.set(i, lang.handleInclude(includes.get(i)));
		}

		// Set the hidden port prefix according to the language
		Links.hiddenPortPrefix = lang.getHiddenPortPrefix();

		switch (compilertype) {
		case LYKOS:
			compilePR();
			break;
		case CH:
			compileCH();
			break;
		case DEFAULT:
			compile();
			break;
		default:
			monitor.add("Please specify a compiler.");
			break;
		}

		monitor.print();
	}

	/**
	 * Compile PR.
	 */
	private void compilePR() {

		ListenerPR listener = new ListenerPR(monitor);
		Interpreter interpreter = new Interpreter(SemanticsType.PR, listener, directories, params, monitor);
		ReoProgram program = interpreter.interpret(files.get(0));

		if (program == null)
			return;

		CompilerSettings settings = new CompilerSettings(files.get(0), Language.JAVA, false);
		settings.ignoreInput(false);
		settings.ignoreData(false);
		settings.partition(!partitioning);
		settings.subtractSyntactically(true);
		settings.commandify(true);
		settings.inferQueues(true);
		settings.put("COUNT_PORTS", false);

		LykosCompiler c = new LykosCompiler(program, files.get(0), outdir, packagename, monitor, settings);

		c.compile();
	}

	/**
	 * Helper class for storing a list of (name, semantics) pairs.
	 *
	 * @param <T> the type of the semantics objects
	 */
	static private class NamedSemantics<T extends Semantics<T>> extends AbstractList<Map.Entry<String,T>> {
		private List<String> nameList = new ArrayList<>();
		private List<T> semanticsList = new ArrayList<>();

		public void add(String name, T semantics) {
			nameList.add(name);
			this.semanticsList.add(semantics);
		}

		@Override
		public void add(int index, Map.Entry<String,T> e)
		{
			nameList.add(index, e.getKey());
			semanticsList.add(index, e.getValue());
		}

		@Override
		public int size()
		{
			return semanticsList.size();
		}

		@Override
		public Map.Entry<String, T> get(int index)
		{
			return new java.util.AbstractMap.SimpleImmutableEntry<>(nameList.get(index), semanticsList.get(index));
		}

		public List<T> getSemanticsList()
		{
			return semanticsList;
		}

		public void applyToAll(Function<Map.Entry<String, T>, T> f)
		{
			ArrayList<T> newSemanticsList = new ArrayList<T>();
			for (int i = 0; i < size(); i++) {
				Map.Entry<String, T> e = get(i);
				T result = f.apply(e);
				semanticsList.set(i, result);
			}
		}
	};

	/**
	 * Default compilation method.
	 */
	private void compile() {

		long t0 = System.nanoTime();
		
		Interpreter interpreter = getInterpreter(lang);

		ReoProgram program;
		if ((program = interpreter.interpret(files.get(0))) == null)
			return;
		
		long t1 = System.nanoTime();
		System.out.println("Interpret        : \t" + (t1-t0)/1E9);
		
		ReoConnector connector = addPortWindows(program.getConnector(), lang);
		connector = connector.propagate(monitor);
		if (debug) System.out.println("\n\nPropagated:\n" + connector);
		connector = connector.generateMissingInstanceNames();
		connector = connector.flatten();
		if (debug) System.out.println("\n\nFlattened:\n" + connector);
		connector = connector.insertNodes(true, false, new RuleBasedAutomaton());
		// generate instance names for the inserted nodes
		connector = connector.generateMissingInstanceNames();
		if (debug) System.out.println("\n\nNoded:\n" + connector);
		connector = connector.integrate();
		if (debug) System.out.println("\n\nIntegrated:\n" + connector);

		List<ReoConnectorAtom> atoms = new LinkedList<>(connector.getAtoms());

		List<Component> components = filterAndBuildAtomics(atoms, lang);

		// do reward monitors here as they are akin to atomics
		// in particular they are treated like atomics for the call
		// to getDualInterface below
		if (lang == Language.PRISM) {
			components.addAll(filterAndBuildRewardMonitors(atoms));
		}

		long t4 = 0;
		if (!dontComposeProtocol) {
			Set<Port> intface = getDualInterface(components);

			NamedSemantics<RuleBasedAutomaton> protocol = getProtocol(atoms, lang, RuleBasedAutomaton.class);
		
			long t2 = System.nanoTime();
			System.out.println("Flatten          : \t" + (t2-t1)/1E9);

			// rename the memory cells
			protocol.applyToAll(this::renameMemoryCells);

			RuleBasedAutomaton composition = new RuleBasedAutomaton().compose(protocol.getSemanticsList(), false);
			if (debug) System.out.println("\n\nComposed:\n" + composition);

			composition = composition.restrict(intface);
			if (debug) System.out.println("\n\nRestricted:\n" + composition);

			long t3 = System.nanoTime();
			System.out.println("Compose & hide   : \t" + (t3-t2)/1E9);

			Set<Transition> transitions = buildTransitions(composition);

			t4 = System.nanoTime();
			System.out.println("Commandification : \t" + (t4-t3)/1E9);

			Set<Set<Transition>> partition = partition(transitions);

			List<Component> protocols = buildProtocols(composition, partition);

			if (protocolFirst) {
				components.addAll(0, protocols);
			} else {
				components.addAll(protocols);
			}
		} else {
			// construct uncomposed protocol
			NamedSemantics<RuleBasedAutomaton> protocol = getProtocol(atoms, lang, RuleBasedAutomaton.class);

					
			long t2 = System.nanoTime();
			System.out.println("Flatten          : \t" + (t2-t1)/1E9);

			List<Component> protocols  = new ArrayList<>();
			for (Map.Entry<String, RuleBasedAutomaton> p : protocol) {
				RuleBasedAutomaton rba = p.getValue();

				// rename the memory cells by prefixing the name
				rba = rba.renameMemory(p.getKey() + "_");

				// compute the transitions
				Set<Transition> rbaTransitions = buildTransitions(rba);

				protocols.add(buildProtocol(rba, p.getKey(), rbaTransitions));
			}
			t4 = System.nanoTime();
			System.out.println("Commandification : \t" + (t4-t2)/1E9);

			if (protocolFirst) {
				components.addAll(0, protocols);
			} else {
				components.addAll(protocols);
			}
		}

		if (!atoms.isEmpty()) {
			monitor.add("Not all components have semantics.");
		}

		ReoTemplate template = new ReoTemplate(program.getFile(), version, packagename, program.getName(), components, includes, settings);
		generateCode(template);

		long t5 = System.nanoTime();
		System.out.println("Template         : \t" + (t5-t4)/1E9);
		System.out.println("Total            : \t" + (t5-t1)/1E9);
	}

	private Interpreter getInterpreter(Language lang) {
		switch (lang) {
		case PRISM:
			ListenerPRBA listenerPRBA = new ListenerPRBA(monitor);
			return new Interpreter(SemanticsType.CH, listenerPRBA, directories, params, monitor);
		case JAVA:
		case C11:
			ListenerRBA listenerRBA = new ListenerRBA(monitor);
			return new Interpreter(SemanticsType.RBA, listenerRBA, directories, params, monitor);
		case MAUDE:
			ListenerRBA listenerRba = new ListenerRBA(monitor);
			return new Interpreter(SemanticsType.RBA, listenerRba, directories, params, monitor);
		case PRT:
			break;
		case TEXT:
			break;
		default:
			break;
		}
		return null;
	}
	
	/**
	 * Compile.
	 */
	private void compileCH() {

		Interpreter interpreter = getInterpreter(lang);

		ReoProgram program;
		if ((program = interpreter.interpret(files.get(0))) == null)
			return;

		ReoConnector connector = addPortWindows(program.getConnector(), lang);
		connector = connector.propagate(monitor);
		connector = connector.generateMissingInstanceNames();
		connector = connector.flatten();
		connector = connector.insertNodes(true, false, new RuleBasedAutomaton());
		// generate instance names for the inserted nodes
		connector = connector.generateMissingInstanceNames();
		connector = connector.integrate();

		List<ReoConnectorAtom> atoms = new LinkedList<>(connector.getAtoms());

		List<Component> components = filterAndBuildAtomics(atoms, lang);

		Set<Port> intface = getDualInterface(components);

		NamedSemantics<RuleBasedAutomaton> protocol = getProtocol(atoms, lang, RuleBasedAutomaton.class);
	
		List<ConstraintHypergraph> ch = new ArrayList<>();
		for(RuleBasedAutomaton rba : protocol.getSemanticsList()){
			ch.add(new ConstraintHypergraph(rba.getAllRules(),rba.getInitial()));
		}
		
		ConstraintHypergraph composition = new ConstraintHypergraph().compose(ch);

		composition = composition.restrict(intface);

		Set<Transition> transitions = buildTransitions(composition);

		Set<Set<Transition>> partition = partition(transitions);

		List<Component> protocols = buildProtocols(composition, partition);

		components.addAll(protocols);


		if (!atoms.isEmpty()) {
			monitor.add("Not all components have semantics.");
		}

		ReoTemplate template = new ReoTemplate(program.getFile(), version, packagename, program.getName(), components, includes, settings);
		generateCode(template);
	}

	/**
	 * Closes a given connector by attaching port windows to visible ports.
	 * 
	 * @param connector
	 *            potentially open connector
	 * @param lang
	 *            target language
	 * @return Connector with all ports hidden.
	 */
	private ReoConnector addPortWindows(ReoConnector connector, Language lang) {
		if (lang != Language.JAVA && lang != Language.MAUDE)
			return connector;
		List<ReoConnector> list = new ArrayList<>();
		list.add(connector);
		for (Port p : connector.getInterface()) {
			if (!p.isHidden()) {
				if(lang == Language.JAVA){
					Value v = new StringValue(p.getName());
					List<Value> values = Arrays.asList(v);
					String call = p.isInput() ? "Windows.producer" : "Windows.consumer";
					Reference ref = new Reference(call, Language.JAVA, new ArrayList<>(), values);
	
					PortType t = p.isInput() ? PortType.OUT : PortType.IN;
					Port q = new Port(p.getName(), t, p.getPrioType(), new TypeTag("String"), true);
					Map<Port, Port> links = new LinkedHashMap<>();
					links.put(q, q);
	
					ReoConnectorAtom window = new ReoConnectorAtom("PortWindow", null, Arrays.asList(ref), links);
					list.add(window);
				}
				if(lang == Language.MAUDE){
					Value v = new StringValue(p.getName());
					List<Value> values = Arrays.asList(v);
					String call = p.isInput() ? "prod" : "cons";
					Reference ref = new Reference(call, Language.MAUDE, new ArrayList<>(), values);
	
					PortType t = p.isInput() ? PortType.OUT : PortType.IN;
					Port q = new Port(p.getName(), t, p.getPrioType(), new TypeTag("String"), true);
					Map<Port, Port> links = new LinkedHashMap<>();
					links.put(q, q);
	
					ReoConnectorAtom window = new ReoConnectorAtom("PortWindow", null, Arrays.asList(ref), links);
					list.add(window);
				}
			}
		}
		return new ReoConnectorComposite(null, connector.getInstanceName(), "", list).rename(new LinkedHashMap<>());
	}

	/**
	 * Collates all atomic component templates, removing them from the list of atoms
	 * and returning corresponding components.
	 * 
	 * @param atoms
	 *            list of atoms; those that where processed are removed from the list
	 * @param lang
	 *            target language
	 * @return list of atomic component templates
	 */
	private List<Component> filterAndBuildAtomics(List<ReoConnectorAtom> atoms, Language lang) {
		List<Component> components = new ArrayList<>();
		int n_atom = 1;
		Iterator<ReoConnectorAtom> it = atoms.iterator();
		while (it.hasNext()) {
			ReoConnectorAtom atom = it.next();
			Reference r = atom.getReference(lang);
			if (r != null) {
				it.remove(); // handled

				String call = r.getCall();
				String name = atom.getInstanceName();

				// TODO the string representation of parameter values is target
				// language dependent.
				List<String> params = new ArrayList<>();
				for (Value v : r.getValues()) {
					if (v instanceof BooleanValue) {
						params.add(((BooleanValue) v).getValue() ? "true" : "false");
					} else if (v instanceof StringValue) {
						params.add("\"" + ((StringValue) v).getValue() + "\"");
					} else if (v instanceof DecimalValue) {
						params.add(Double.toString(((DecimalValue) v).getValue()));
					}
				}

				components.add(new Atomic(name, params, atom.getInterface(), call, atom.getLinks()));
			}
		}
		return components;
	}

	/**
	 * Gets the dual of the interface from atomic components to protocol
	 * components.
	 * 
	 * @param atomics
	 *            list of atomic components
	 * @return Dual interface
	 */
	private Set<Port> getDualInterface(List<Component> atomics) {
		Set<Port> intface = new LinkedHashSet<>();
		for (Component atom : atomics) {
			for (Port p : atom.getPorts()) {
				PortType t = p.isInput() ? PortType.OUT : PortType.IN;
				intface.add(new Port(p.getName(), t, p.getPrioType(), p.getTypeTag(), true));
			}
		}
		return intface;
	}

	/**
	 * Composes all protocol connector in a given connector. The protocol
	 * consists of all components without a reference to code in a given target
	 * language.
	 * 
	 * @param connector
	 *            connector
	 * @param lang
	 *            target language
	 * @param type
	 *            type
	 * @return List of all protocol connectors.
	 */
	private <T extends Semantics<T>> NamedSemantics<T> getProtocol(List<ReoConnectorAtom> atoms, Language lang, Class<T> type) {
		NamedSemantics<T> result = new NamedSemantics<>();

		Iterator<ReoConnectorAtom> it = atoms.iterator();
		while (it.hasNext()) {
			ReoConnectorAtom atom = it.next();

			Reference r = atom.getReference(lang);
			if (r == null) {
				T semantics = null;
				for (Atom x : atom.getSemantics())
					if (type.isInstance(x))
						semantics = type.cast(x);
				if (semantics == null) {
					// skip
					continue;
				}

				it.remove();
				result.add(atom.getInstanceName(), semantics);
			}
		}

		return result;
	}

	private RuleBasedAutomaton renameMemoryCells(Map.Entry<String, RuleBasedAutomaton> e) {
		RuleBasedAutomaton rba = e.getValue();
		return rba.renameMemory(e.getKey() + "_");
	}

	private List<Component> filterAndBuildRewardMonitors(List<ReoConnectorAtom> atoms) {
		LinkedHashMap<String, RewardMonitor> globalMonitors = new LinkedHashMap<>();
		List<Component> components = new ArrayList<>();
		Iterator<ReoConnectorAtom> it = atoms.iterator();
		while (it.hasNext()) {
			ReoConnectorAtom atom = it.next();
			RewardMonitor rewMonitor = (RewardMonitor)atom.getSemantics(SemanticsType.REWARD);
			if (rewMonitor == null)
				continue;

			it.remove();

			// perform renaming
			rewMonitor = rewMonitor.rename(atom.getLinks());
			
			String globalName = rewMonitor.getGlobalName();
			if (globalName != null) {
				globalMonitors.merge(globalName, rewMonitor, (a,b) -> {return a.compose(b);});
			} else {
				components.add(new RewardsAtomic(atom.getInstanceName(), rewMonitor)); 
			}
		}

		// handle global
		for (Entry<String, RewardMonitor> e : globalMonitors.entrySet()) {
			components.add(new RewardsAtomic(e.getKey(), e.getValue()));
		}

		return components;
	}
	
	private Set<Transition> buildTransitions(ConstraintHypergraph protocol) {
		Set<Transition> transitions = new LinkedHashSet<>();
		for (Rule rule : protocol.getRules()){
			Command cmd = Commands.commandify(rule.getFormula());
			transitions.add(cmd.toTransition(lang));
		}
		return transitions;
	}
	
	private Set<Transition> buildTransitions(RuleBasedAutomaton protocol) {
		Set<Transition> transitions = new LinkedHashSet<>();
		for (Rule rule : protocol.getAllRules()) {
			Command cmd = Commands.commandify(rule.getFormula());
			transitions.add(cmd.toTransition(lang));
		}
		return transitions;
	}

	/**
	 * Build the protocol templates from a composed RBA and a partition (set of set of transitions).
	 * 
	 * @param protocol
	 *            composed protocol
	 * @param partition
	 *            partitioned set of transitions
	 * 
	 * @return list of protocol templates
	 */
	private List<Component> buildProtocols(RuleBasedAutomaton protocol, Set<Set<Transition>> partition) {
		List<Component> components = new ArrayList<>();
		int n_protocol = 1;
		for (Set<Transition> part : partition) {
			components.add(buildProtocol(protocol, "Protocol" + n_protocol++, part));
		}
		return components;
	}

	/**
	 * Build the single protocol template from an RBA and a set of transitions.
	 * 
	 * @param protocol
	 *            the protocol RBA
	 * @param name
	 *            the name of the protocol
	 * @param transitions
	 *            set of transitions for the RBA
	 * 
	 * @return list of protocol templates
	 */
	private Protocol buildProtocol(RuleBasedAutomaton protocol, String name, Set<Transition> transitions) {
		// Get the interface of this part
		Set<Port> ports = new LinkedHashSet<>();
		for (Transition t : transitions)
			ports.addAll(t.getInterface());

		// Find the type of each memory cell that occurs in this part
		Map<String, TypeTag> tags = new LinkedHashMap<>();
		for (Map.Entry<MemoryVariable, Term> init : protocol.getInitial().entrySet()) {
			tags.put(init.getKey().getName(), init.getValue().getTypeTag());
		}

		for (Transition t : transitions) {
			for (Map.Entry<MemoryVariable, Term> upd : t.getMemory().entrySet()) {
				if (!tags.containsKey(upd.getKey().getName())) {
					tags.put(upd.getKey().getName(), upd.getValue().getTypeTag());
				}
			}
		}

		// Get the initial value of each memory cell
		Map<String, Term> initial = new HashMap<>();
		for (Map.Entry<MemoryVariable, Term> e : protocol.getInitial().entrySet()) {
			initial.put(e.getKey().getName(), e.getValue());
		}

		SortedMap<MemoryVariable, Object> protocolInitial = new TreeMap<>();
		for (Map.Entry<String, Term> e : initial.entrySet()) {
			protocolInitial.put(new MemoryVariable(e.getKey(), false), e.getValue());
		}

		for (Map.Entry<String, TypeTag> e : tags.entrySet()) {
			if (!initial.containsKey(e.getKey())) {
				protocolInitial.put(new MemoryVariable(e.getKey(), false), null);
			}
		}

		return new Protocol(name, ports, transitions, protocolInitial);
	}

	/**
	 * Build the protocol templates from a constraint hypergraph and a
	 * partition.
	 * 
	 * @param protocol
	 *            composed protocol
	 * @param partition
	 *            partitioned set of transitions
	 * 
	 * @return list of protocol templates
	 */
	private List<Component> buildProtocols(ConstraintHypergraph protocol, Set<Set<Transition>> partition) {
		List<Component> components = new ArrayList<>();
		int n_protocol = 1;
		for (Set<Transition> part : partition) {

			// Get the interface of this part
			Set<Port> ports = new LinkedHashSet<>();
			for (Transition t : part)
				ports.addAll(t.getInterface());

			// Find the type of each memory cell that occurs in this part
			Map<String, TypeTag> tags = new LinkedHashMap<>();
			for (Map.Entry<MemoryVariable, Term> init : protocol.getInitials().entrySet())
				tags.put(init.getKey().getName(), init.getValue().getTypeTag());
			for (Transition t : part)
				for (Map.Entry<MemoryVariable, Term> upd : t.getMemory().entrySet())
					tags.put(upd.getKey().getName(), upd.getValue().getTypeTag());

			// Get the initial value of each memory cell
			Map<MemoryVariable, Object> initial = new LinkedHashMap<>();
			for (Map.Entry<String, TypeTag> e : tags.entrySet())
				initial.put(new MemoryVariable(e.getKey(), false, e.getValue()), protocol.getInitials().get(new MemoryVariable(e.getKey(),false,e.getValue())));

			if(lang == Language.JAVA)
				components.add(new Protocol("Protocol" + n_protocol++, ports, part, initial));
			if(lang == Language.MAUDE)
				components.add(new MaudeProtocol("Protocol" + n_protocol++, ports, part, initial));
		}
		return components;
	}

	/**
	 * Partitions the set of transitions into independent regions.
	 * 
	 * @param transitions
	 *            set of transitions
	 * @return partitioned set of transitions.
	 */
	private Set<Set<Transition>> partition(Set<Transition> transitions) {
		Set<Set<Transition>> partition = new LinkedHashSet<>();

		if (!transitions.isEmpty()) {
			// for (Transition t : transitions) {
			// Set<Set<Transition>> newPartition = new LinkedHashSet<>();
			// newPartition.add(new LinkedHashSet<>(Arrays.asList(t)));
			// for (Set<Transition> part : partition) {
			//
			// }
			// }
			partition.add(transitions);
		}

		return partition;
	}

	/**
	 * Generates code from the standard Reo template.
	 * 
	 * @param template
	 *            template with code
	 */
	private void generateCode(ReoTemplate template) {
		STGroup group = null;
		String extension = "";

		switch (lang) {
		case JAVA:
			group = new STGroupFile("Java.stg");
			extension = ".java";
			break;
		case MAUDE:
			group = new STGroupFile("Maude.stg");
			extension = ".maude";
			break;
		case PRISM:
			group = new STGroupFile("Prism.stg");
			extension = ".prism";
			break;
		default:
			return;
		}

		ST stringtemplate = group.getInstanceOf("main");
		stringtemplate.add("S", template);

		String code = stringtemplate.render(72);

		try {
			File file = new File(outdir + File.separator + template.getName() + extension);
			file.getParentFile().mkdirs();
			FileWriter out = new FileWriter(file);
			out.write(code);
			out.close();
		} catch (IOException e) {
		}
	}
}
