package nl.cwi.reo.interpret.connectors;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.Map;

// TODO: Auto-generated Javadoc
/**
 * Enumeration of all compiler supported languages.
 */
public enum Language {

	/**
	 * Java language.
	 */
	JAVA,

	/** C11 language. */
	C11,

	/** Maude language. */
	MAUDE,

	/** PRISM language for probabilistic model checking. */
	PRISM {
		@Override
		public String getHiddenPortPrefix() {
			return "_";
		}

		@Override
		public String handleInclude(String include) throws RuntimeException {
			try {
				String text = new String(Files.readAllBytes(Paths.get(include)), StandardCharsets.UTF_8);
				return
						"\n// ---- included from " + include + " --------\n" + 
						text + 
						"\n// --------------------------------------\n\n";
			} catch (java.nio.file.NoSuchFileException e) {
				throw new RuntimeException("File not found: " + include);
			} catch (IOException e) {
				throw new RuntimeException(e.toString());
			}
		}

		
		@Override
		public void setDefaults(Map<String, Object> settings) {
			settings.putIfAbsent("PRISM_MODEL", "mdp");
		}
		
		public void validateSettings(Map<String, Object> settings) throws RuntimeException {
			if (!(settings.get("PRISM_MODEL") instanceof String) ||
				!settings.get("PRISM_MODEL").toString().matches("[a-z]+")) {
				throw new RuntimeException("Value for PRISM_MODEL has to be identifier, got '" + settings.get("PRISM_MODEL") + "'");
			}
		}

		@Override
		public String getSettingsUsage() {
			return
					" * PRISM_MODEL=<x>       specify PRISM model type (default: mdp)\n" + 
					" * PRISM_DATA_ABSTRACT   generate data-abstract Reo circuit";
		}
	},

	/** Proto Runtime. */
	PRT,

	/** Pr text. */
	TEXT;

	/** Get the hidden port prefix appropriate for this language */
	public String getHiddenPortPrefix() {
		return "$";
	}

	public void setDefaults(Map<String, Object> settings) {
		// NOP
	}

	public String getSettingsUsage() {
		return "";
	}

	public void validateSettings(Map<String, Object> settings) throws RuntimeException {
		// NOP
	}

	public String handleInclude(String include) {
		return include;
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public String toString() {
		switch (this) {
		case JAVA:
			return "Java";
		case C11:
			return "C11";
		case MAUDE:
			return "Maude";
		case PRISM:
			return "PRISM";
		case PRT:
			return "Proto runtime";
		case TEXT:
			return "Text";
		default:
			throw new IllegalArgumentException();
		}
	}
}
