package nl.cwi.reo.semantics.prba;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.LinkedHashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.checkerframework.checker.nullness.qual.Nullable;

import nl.cwi.reo.interpret.Atom;
import nl.cwi.reo.interpret.Scope;
import nl.cwi.reo.interpret.SemanticsType;
import nl.cwi.reo.interpret.ports.Port;
import nl.cwi.reo.semantics.Semantics;
import nl.cwi.reo.util.Monitor;

/**
 * A reward monitor.
 */
public final class RewardMonitor implements Semantics<RewardMonitor> {

	/** A rule of the monitor */
	public final static class Rule {
		/** The list of ports that have to be simultaneously active to earn the reward */
		private List<Port> ports;
		/** The reward value */
		private int value;
		
		/** Constructor */
		public Rule(List<Port> ports, int value) {
			this.ports = Collections.unmodifiableList(ports);
			this.value = value;
		}

		/** Get the ports */
		public List<Port> getPorts()
		{
			return ports;
		}

		/** Get the reward value */
		public int getValue()
		{
			return value;
		}

		/** Return rule with the given renaming */
		public Rule rename(Map<Port, Port> links) {
			List<Port> _ports = new ArrayList<>();
			for (Port port : ports) {
				Port p = links.get(port);
				if (p == null)
					p = port;
				_ports.add(p);
			}
			return new Rule(_ports, value);
		}

		@Override
		public String toString()
		{
			StringBuilder sb = new StringBuilder();
			boolean first = true;
			for (Port p : ports) {
				if (!first) sb.append(", ");
				first = false;
				sb.append(p);
			}
			sb.append(" : ");
			sb.append(value);
			sb.append(";\n");
			return sb.toString();
		}
	}

	/**
	 * The global name (for a global, shared reward structure).
	 */
	private @Nullable String globalName;

	/**
	 * The list of rules.
	 */
	private List<Rule> rules;

	/** Constructor */
	public RewardMonitor(@Nullable String globalName, List<Rule> rules) {
		this.globalName = globalName;
		this.rules = rules;
	}

	@Override
	public SemanticsType getType()
	{
		return SemanticsType.REWARD;
	}

	@Override
	public Set<Port> getInterface()
	{
		Set<Port> intface = new LinkedHashSet<>();
		for (Rule rule : rules)
			intface.addAll(rule.getPorts());
		return intface;
	}

	@Override
	public RewardMonitor rename(Map<Port, Port> links)
	{
		List<Rule> _rules = new ArrayList<>();
		for (Rule rule: rules) {
			_rules.add(rule.rename(links));
		}

		return new RewardMonitor(globalName, _rules);
	}

	@Override
	public Atom getNode(Set<Port> node)
	{
		throw new UnsupportedOperationException();
	}

	@Override
	public @Nullable Atom evaluate(Scope s, Monitor m)
	{
		return this;
	}

	@Override
	public RewardMonitor getDefault(Set<Port> iface)
	{
		List<Port> ports = new ArrayList<>();
		ports.addAll(iface);
		return new RewardMonitor(globalName, Collections.singletonList(new Rule(ports, 0)));
	}

	public RewardMonitor compose(RewardMonitor other)
	{
		List<Rule> _rules = new ArrayList<>();
		_rules.addAll(rules);
		_rules.addAll(other.getRules());
		return new RewardMonitor(globalName, _rules);
	}

	@Override
	public RewardMonitor compose(List<RewardMonitor> components)
	{
		List<Rule> _rules = new ArrayList<>();
		_rules.addAll(rules);
		for (RewardMonitor r : components) {
			_rules.addAll(r.getRules());
		}
		return new RewardMonitor(globalName, _rules);
	}
	
	/** Get the global name (if there is one). */
	@Nullable
	public String getGlobalName()
	{
		return globalName;
	}

	/** Get the rules. */
	public List<Rule> getRules()
	{
		return rules;
	}

	@Override
	public RewardMonitor restrict(Collection<? extends Port> intface)
	{
		// ignore
		return this;
	}

	@Override
	public String toString()
	{
		StringBuffer sb = new StringBuffer();
		sb.append("#REWARD");
		if (globalName != null) {
			sb.append(" " + globalName);
		}
		sb.append("\n");
		for (Rule rule : rules) {
			sb.append(rule);
		}
		return sb.toString();
	}
}